@@ -1,13 +1,14 @@
-#! /usr/bin/python
+#!/usr/bin/python
 #
-# pgdb.py
+# PyGreSQL - a Python interface for the PostgreSQL database.
 #
-# Written by D'Arcy J.M. Cain
+# This file contains the DB-API 2 compatible pgdb module.
 #
-# $Id: pgdb.py 901 2017-01-06 12:25:02Z cito $
+# Copyright (c) 2020 by the PyGreSQL Development Team
 #
+# Please see the LICENSE.TXT file for specific restrictions.
 
-"""pgdb - DB-API 2.0 compliant module for PygreSQL.
+"""pgdb - DB-API 2.0 compliant module for PyGreSQL.
 
 (c) 1999, Pascal Andre <andre@via.ecp.fr>.
 See package documentation for further information on copyright.
@@ -65,7 +66,22 @@ Basic usage:
 
 from __future__ import print_function, division
 
-from _pg import *
+try:
+    from _pg import *
+except ImportError:
+    import os
+    import sys
+    # see https://docs.python.org/3/whatsnew/3.8.html#ctypes
+    if os.name == 'nt' and sys.version_info >= (3, 8):
+        for path in os.environ["PATH"].split(os.pathsep):
+            if os.path.exists(os.path.join(path, 'libpq.dll')):
+                with os.add_dll_directory(os.path.abspath(path)):
+                    from _pg import *
+                break
+        else:
+            raise
+    else:
+        raise
 
 __version__ = version
 
@@ -74,23 +90,27 @@ from time import localtime
 from decimal import Decimal
 from uuid import UUID as Uuid
 from math import isnan, isinf
-from collections import namedtuple, Iterable
+try:
+    from collections.abc import Iterable
+except ImportError:  # Python < 3.3
+    from collections import Iterable
+from collections import namedtuple
 from keyword import iskeyword
 from functools import partial
 from re import compile as regex
 from json import loads as jsondecode, dumps as jsonencode
 
-try:
+try:  # noinspection PyUnresolvedReferences
     long
 except NameError:  # Python >= 3.0
     long = int
 
-try:
+try:  # noinspection PyUnresolvedReferences
     unicode
 except NameError:  # Python >= 3.0
     unicode = str
 
-try:
+try:  # noinspection PyUnresolvedReferences
     basestring
 except NameError:  # Python >= 3.0
     basestring = (str, bytes)
@@ -493,7 +513,7 @@ class Typecasts(dict):
 
     The cast functions get passed the string representation of a value in
     the database which they need to convert to a Python object.  The
-    passed string will never be None since NULL values are already be
+    passed string will never be None since NULL values are already
     handled before the cast function is called.
     """
 
@@ -719,11 +739,11 @@ class TypeCache(dict):
             # older remote databases (not officially supported)
             self._query_pg_type = ("SELECT oid, typname,"
                 " typlen, typtype, null as typcategory, typdelim, typrelid"
-                " FROM pg_type WHERE oid=%s")
+                " FROM pg_catalog.pg_type WHERE oid OPERATOR(pg_catalog.=) %s")
         else:
             self._query_pg_type = ("SELECT oid, typname,"
                 " typlen, typtype, typcategory, typdelim, typrelid"
-                " FROM pg_type WHERE oid=%s")
+                " FROM pg_catalog.pg_type WHERE oid OPERATOR(pg_catalog.=) %s")
 
     def __missing__(self, key):
         """Get the type info from the database if it is not cached."""
@@ -763,7 +783,9 @@ class TypeCache(dict):
         if not typ.relid:
             return None  # this type is not composite
         self._src.execute("SELECT attname, atttypid"
-            " FROM pg_attribute WHERE attrelid=%s AND attnum>0"
+            " FROM pg_catalog.pg_attribute"
+            " WHERE attrelid OPERATOR(pg_catalog.=) %s"
+            " AND attnum OPERATOR(pg_catalog.>) 0"
             " AND NOT attisdropped ORDER BY attnum" % (typ.relid,))
         return [FieldInfo(name, self.get(int(oid)))
             for name, oid in self._src.fetch(-1)]
@@ -1040,14 +1062,15 @@ class Cursor(object):
         rowcount = 0
         sql = "BEGIN"
         try:
-            if not self._dbcnx._tnx:
+            if not self._dbcnx._tnx and not self._dbcnx.autocommit:
                 try:
                     self._src.execute(sql)
                 except DatabaseError:
                     raise  # database provides error message
                 except Exception:
                     raise _op_error("Can't start transaction")
-                self._dbcnx._tnx = True
+                else:
+                    self._dbcnx._tnx = True
             for parameters in seq_of_parameters:
                 sql = operation
                 sql = self._quoteparams(sql, parameters)
@@ -1212,7 +1235,7 @@ class Cursor(object):
         if not table or not isinstance(table, basestring):
             raise TypeError("Need a table to copy to")
         if table.lower().startswith('select'):
-                raise ValueError("Must specify a table, not a query")
+            raise ValueError("Must specify a table, not a query")
         else:
             table = '"%s"' % (table,)
         operation = ['copy %s' % (table,)]
@@ -1220,7 +1243,7 @@ class Cursor(object):
         params = []
         if format is not None:
             if not isinstance(format, basestring):
-                raise TypeError("The frmat option must be be a string")
+                raise TypeError("The format option must be be a string")
             if format not in ('text', 'csv', 'binary'):
                 raise ValueError("Invalid format")
             options.append('format %s' % (format,))
@@ -1428,7 +1451,6 @@ class Cursor(object):
             return _row_factory(tuple(names))
 
 
-
 CursorDescription = namedtuple('CursorDescription',
     ['name', 'type_code', 'display_size', 'internal_size',
      'precision', 'scale', 'null_ok'])
@@ -1457,6 +1479,7 @@ class Connection(object):
         self._tnx = False  # transaction state
         self.type_cache = TypeCache(cnx)
         self.cursor_type = Cursor
+        self.autocommit = False
         try:
             self._cnx.source()
         except Exception:
@@ -1466,7 +1489,18 @@ class Connection(object):
         """Enter the runtime context for the connection object.
 
         The runtime context can be used for running transactions.
+
+        This also starts a transaction in autocommit mode.
         """
+        if self.autocommit:
+            try:
+                self._cnx.source().execute("BEGIN")
+            except DatabaseError:
+                raise  # database provides error message
+            except Exception:
+                raise _op_error("Can't start transaction")
+            else:
+                self._tnx = True
         return self
 
     def __exit__(self, et, ev, tb):
@@ -1508,9 +1542,9 @@ class Connection(object):
                 try:
                     self._cnx.source().execute("COMMIT")
                 except DatabaseError:
-                    raise
+                    raise  # database provides error message
                 except Exception:
-                    raise _op_error("Can't commit")
+                    raise _op_error("Can't commit transaction")
         else:
             raise _op_error("Connection has been closed")
 
@@ -1522,9 +1556,9 @@ class Connection(object):
                 try:
                     self._cnx.source().execute("ROLLBACK")
                 except DatabaseError:
-                    raise
+                    raise  # database provides error message
                 except Exception:
-                    raise _op_error("Can't rollback")
+                    raise _op_error("Can't rollback transaction")
         else:
             raise _op_error("Connection has been closed")
 
@@ -1763,7 +1797,7 @@ class Binary(bytes):
 # Additional type helpers for PyGreSQL:
 
 def Interval(days, hours=0, minutes=0, seconds=0, microseconds=0):
-    """Construct an object holding a time inverval value."""
+    """Construct an object holding a time interval value."""
     return timedelta(days, hours=hours, minutes=minutes, seconds=seconds,
         microseconds=microseconds)
 
@@ -1819,6 +1853,7 @@ class Literal:
 
     __pg_repr__ = __str__
 
+
 # If run as script, print some information:
 
 if __name__ == '__main__':
